#!/usr/bin/env python

#Written by Philipp Birken and Jonathan Bull
#The routines implemented here perform a discrete Fourier analysis
#for the linearized Euler equations, discretized using the JST flux
#and then solved using a preconditioned pseudo time iteration
#This code is licensed under the GPL 3.0

import sys
from math import pi
from math import pow
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.pylab as pylab

def get_scheme_params(args):
  '''Get numerical scheme parameters from integer input arguments.'''

  # RK schemes
  rkscheme_dict = {
    0: 'ERK3',  # 3-stage explicit Runge-Kutta
    1: 'ARK3J', # Jameson 3-stage additive explicit Runge-Kutta
    2: 'ERK51', # 5-stage 1st-order explicit Runge-Kutta
    3: 'ARK51', # van Leer 5-stage 1st-order additive explicit Runge-Kutta
    4: 'ARK52', # van Leer 5-stage 2nd-order additive explicit Runge-Kutta
    5: 'ARK5J'  # Jameson 5-stage 1st-order additive explicit Runge-Kutta
  }
  try:
    rkscheme = rkscheme_dict[int(args[1])]
  except:
    print('Could not parse RK scheme arg, exiting')
    sys.exit()

  # preconditioners
  precond_dict = {
    0: 'LUSGSE', # standard SGS via explicit method (default?)
    1: 'EE',     # exact Jacobian via explicit method
    2: 'SSOR',   # SSOR
    3: 'LUSGSW', # LUSGS for Rosenbrock and W methods
    4: 'EW'      # exact Jacobian for Rosenbrock and W methods
  }
  try:
    precond = precond_dict[int(args[2])]
  except:
    print('Could not parse preconditioner arg, exiting')
    sys.exit()

  # equation choices
  eqn_dict = {
    0: 'steady',
    1: 'unsteady'
  }
  try:
    equation = eqn_dict[int(args[3])]
  except:
    print('Could not parse unsteady arg, exiting')
    sys.exit()

  return rkscheme, precond, equation

def rkcoeffs(scheme):
  bcoeffs = np.empty([0])
  if scheme=='ERK3':
    acoeffs = np.array([0.1481,0.4,1.])
  elif scheme=='ERK51':
    #ERK5 van Leer et. al. 1st order
    acoeffs = np.array([0.0533,0.1263,0.2375,0.4414,1.])
  elif scheme=='ERK52':
    #ERK5 coeffs van Leer et. al. 2nd order
    acoeffs = np.array([0.0695,0.1602,0.2898,0.5060,1.])
  elif scheme=='ARK3':
    #ARK3 coeffs, van Leer et. al. 1st order
    acoeffs = np.array([0.1481,0.4,1.])
    bcoeffs = np.array([1.,0.5,0.5])
  elif scheme=='ARK51':
    #ARK51 van Leer et. al. 1st order
    acoeffs = np.array([0.0533,0.1263,0.2375,0.4414,1.])
    bcoeffs = np.array([1.,0.,0.56,0.,0.44])
  elif scheme=='ARK52':
    #ARK52 van Leer et. al. 2nd order
    acoeffs = np.array([0.0695,0.1602,0.2898,0.5060,1.])
    bcoeffs = np.array([1.,0.,0.56,0.,0.44])
  elif scheme=='ARK5J':
    #ARK5J Jameson
    acoeffs = np.array([0.25,1./6.,0.375,0.5,1.])
    bcoeffs = np.array([1.,0.,0.56,0.,0.44])
  else:
    print('RK scheme not found. Exiting.')
    sys.exit()
  
  return acoeffs,bcoeffs

# use numpy to calculate numerical amplification factors
def amp_factor_numeric(Nx,Ny,k_i,k_j,scheme,precond,unsteady,x0):

  # primitive variables
  r = 1.
  #flow angle, e.g. 0 or np.pi/4
  alpha = 0.
  u = np.cos(alpha)
  v = np.sin(alpha)
  # normal velocity
  un = np.sqrt(u**2+v**2)
  # Mach
  M = 0.8
  # speed of sound
  c = un/M
  # gamma
  g = 1.4
  # pressure
  p = r*c**2/g
  # energy
  E = p/(r*(g-1))+un**2/2.
  # enthalpy
  H = E+p/r
  
  # numerical scheme coeffs
  acoeffs,bcoeffs = rkcoeffs(scheme)
  additive = 1
  if(len(bcoeffs)==0):
    additive = 0

  # input args
  CFLstar=x0[0]
  #CFLstar = 100
  dlim=x0[1]
  #dlim = .5
  eps=x0[2]
  #eps = 0.8
  dx = 1./Nx
  dy = 1./Ny
  CFL = 200
  dt = CFL*min(dx,dy)/(un+c)
  dtstar = CFLstar*min(dx,dy)/(un+c)
  V = dx*dy

  # adjusted convection coeffs
  acoeffs = acoeffs*dtstar

  # identity matrix
  Id = np.eye(4)

  # shift operators in x direction
  Exp1 = np.exp(1j*2.*np.pi*k_i/Nx)
  Exp2 = Exp1**2
  Exm1 = np.exp(-1j*2.*np.pi*k_i/Nx)
  Exm2 = Exm1**2
  Cxdiff = Exp1-2+Exm1
  Cxsum  = Exp1+2+Exm1
      
  # shift operators in y direction
  Eyp1 = np.exp(1j*2.*np.pi*k_j/Ny)
  Eyp2 = Eyp1**2
  Eym1 = np.exp(-1j*2.*np.pi*k_j/Ny)
  Eym2 = Eym1**2
  Cydiff = Eyp1-2+Eym1
  Cysum  = Eyp1+2+Eym1
      
  # Spectral radius correction coeffs
  rxcoeff = abs(u)+c
  rycoeff = abs(v)+c
  rcorrect = pow(rycoeff/rxcoeff,2./3.)
  rxcoeff = rxcoeff*(1.+rcorrect)
  rycoeff = rycoeff*(1.+1./rcorrect)
  #Pressure sensor
  sxcoeff = (g-1)*(Cxdiff*r*E-0.5*un*un*Cxdiff*r)/((g-1)*(Cxsum*r*E-0.5*un*un*Cxsum*r)+0.001)
  sycoeff = (g-1)*(Cydiff*r*E-0.5*un*un*Cydiff*r)/((g-1)*(Cysum*r*E-2.*un*un*Cysum*r)+0.001)
  sxcoeff = min(0.25,abs(sxcoeff))
  sycoeff = min(0.25,abs(sycoeff))
  # JST coeffs in smooth flow
  ex2 = abs(sxcoeff*rxcoeff)
  ey2 = abs(sycoeff*rycoeff)
  ex4 = max(0,rxcoeff/32.-2.*ex2)
  ey4 = max(0,rycoeff/32.-2.*ey2)
  
  # conservative form of advective matrices in x and y directions
  A = np.array([[0.,1.,0.,0.],
              [-u**2+(g-1.)/2.*un**2,(3.-g)*u,-(g-1.)*v,(g-1.)],
              [-u*v,v,u,0.],
              [-u*(g*E-(g-1.)*un**2),g*E-(g-1.)/2.*(un**2+2.*u**2),-(g-1.)*u*v,g*u]])
  B = np.array([[0.,0.,1.,0.],
              [-u*v,v,u,0.],
              [-v**2+(g-1.)/2.*un**2,-(g-1.)*u,(3.-g)*v,(g-1.)],
              [-v*(g*E-(g-1.)*un**2),-(g-1.)*u*v,g*E-(g-1.)/2.*(un**2+2.*v**2),g*v]])

  # split A,B into +ve and -ve parts via eigenvalues and entropy fix
  EA,VA = np.linalg.eig(A)
  EB,VB = np.linalg.eig(B)
  EAp = np.zeros(4)
  EAm = np.zeros(4)
  EBp = np.zeros(4)
  EBm = np.zeros(4)
  # entropy fix with dlim parameter
  a = dlim*c
  for i in range(4):
    # A matrix
    e = abs(EA[i])
    if e < a:
      e = 0.5*(a+e*e/a)
    EAp[i] = 0.5*(EA[i] + e)
    EAm[i] = 0.5*(EA[i] - e)
    # B matrix
    e = abs(EB[i])
    if e < a:
      e = 0.5*(a+e*e/a)
    EBp[i] = 0.5*(EB[i] + e)
    EBm[i] = 0.5*(EB[i] - e)
  Ap = np.dot(np.dot(VA,np.diag(EAp)),np.linalg.inv(VA))
  Am = np.dot(np.dot(VA,np.diag(EAm)),np.linalg.inv(VA))
  Bp = np.dot(np.dot(VB,np.diag(EBp)),np.linalg.inv(VB))
  Bm = np.dot(np.dot(VB,np.diag(EBm)),np.linalg.inv(VB))
      
  # steady advective residual operator
  Lc = (dy/2.*A*(Exp1-Exm1) + dx/2.*B*(Eyp1-Eym1))/V

  # first order upwind
  #Lc = (dy*A*(1.-Exm1) + dx*B*(1.-Eym1))/V
  #Preconditioner discretization
  #Lc = 1./V*(dy*((Ap-Am) - Ap*Exm1 + Am*Exp1) + dx*((Bp-Bm) - Bp*Eym1 + Bm*Eyp1))

  # dissipative residual operator: JST artificial viscosity
  Ld = (((2.-Exp1-Exm1)*ex2*dy+(2.-Eyp1-Eym1)*ey2*dx) + ((6.+Exm2+Exp2-4.*Exm1-4.*Exp1)*ex4*dy+(6.+Eym2+Eyp2-4.*Eym1-4.*Eyp1)*dx*ey4))/V
  Dhat = np.array([[1.,0.,0.,0.],
                   [0.,1.,0.,0.],
                   [0.,0.,1.,0.],
                   [-0.5*(g-1)*un*un,0.,0.,g]])
  Ld = Dhat*Ld

  # adjust convective operator if unsteady
  if(unsteady):
    Lc += 3./(2.*dt)*Id

  # choose preconditioning method and compute P matrix
  if(precond=='LUSGSE'):
    # SGS via the explicit method
    D = 1./V*(dy*(Ap-Am)+dx*(Bp-Bm))
    if(unsteady):
      D +=  Id*3./2./dt
    # Rossow L^+, or standard form (D+L)
    L = - 1./V*(dy*Ap*Exm1 + dx*Bp*Eym1)
    # Rossow L^-, or standard form (D+U)
    U = 1./V*(dy*Am*Exp1 + dx*Bm*Eyp1)
    P = np.dot(np.linalg.inv(D+L),np.dot(D,np.linalg.inv(D+U)))

  if(precond=='EE'):
    # Exact Jacobian via the explicit method
    D = 1./V*(dy*(Ap-Am)+dx*(Bp-Bm))
    if(unsteady):
      D +=  Id*3./2./dt
    # Rossow L^+, or standard form (D+L)
    L = - 1./V*(dy*Ap*Exm1 + dx*Bp*Eym1)
    # Rossow L^-, or standard form (D+U)
    U = 1./V*(dy*Am*Exp1 + dx*Bm*Eyp1)
    #Exact Jacobian as preconditioner
    P = D+L+U
    P=np.linalg.inv(P)
      
  if(precond=='SSOR'):
    # the SSOR preconditioner
    omega = 0.9
    D = 1./V*(dy*(Ap-Am)+dx*(Bp-Bm))
    if(unsteady):
      D += Id*3./2./dt
    # Rossow L^+, or standard form (D+L)
    L = D - omega/V*(dy*Ap*Exm1 + dx*Bp*Eym1)
    # Rossow L^-, or standard form (D+U)
    U = D + omega/V*(dy*Am*Exp1 + dx*Bm*Eyp1)
    P = omega*(2.-omega)*np.dot(np.linalg.inv(L),np.dot(D,np.linalg.inv(U)))

  if(precond=='LUSGSW'):
    #The Rosenbrock and W methods
    #  LUSGS preconditioner
    D = Id+eps*dtstar/V*(dy*(Ap-Am)+dx*(Bp-Bm))
    if(unsteady):
      D += Id*eps*3.*dtstar/2./dt
    # Rossow L^+, or standard form (D+L)
    L = D - eps*dtstar/V*(dy*Ap*Exm1 + dx*Bp*Eym1)
    # Rossow L^-, or standard form (D+U)
    U = D + eps*dtstar/V*(dy*Am*Exp1 + dx*Bm*Eyp1)
    # Jameson version: one forward-backward sweep
    P = np.dot(np.linalg.inv(L),np.dot(D,np.linalg.inv(U)))

  if(precond=='EW'):
    #The Rosenbrock and W methods
    # the exact inverse as preconditioner
    D = Id+eps*dtstar/V*(dy*(Ap-Am)+dx*(Bp-Bm))
    if(unsteady):
      D += Id*eps*3.*dtstar/2./dt
    L = - eps*dtstar/V*(dy*Ap*Exm1 + dx*Bp*Eym1)
    U = eps*dtstar/V*(dy*Am*Exp1 + dx*Bm*Eyp1)
    P = D+L+U
    P=np.linalg.inv(P)

  # apply preconditioner P
  Lc = np.dot(P,Lc)
  Ld = np.dot(P,Ld)

  # now the amplification factors
  # ARK schemes
  if(additive):
    fv_old = Ld
    k = Id - acoeffs[0]*(Lc+fv_old)
    for i in range(1,len(acoeffs)):
      a = acoeffs[i]
      b = bcoeffs[i]
      fc = np.dot(Lc,k)
      fv = b*np.dot(Ld,k) + (1.-b)*fv_old
      k = Id - a*(fc+fv)
      fv_old = fv
    G = k
  # ERK schemes
  else:
    k = Id
    for a in acoeffs:
      f = np.dot(k,(Lc+Ld))
      k = Id - a*f
    G = k
  
  return G

def discrete_fourier_analysis(x0,Ly,scheme,precond,equation):
  '''Runs discrete Fourier analysis in pi*pi wavenumber space.
  Plots eigenvalues in complex plane and smoothing factor contours.
  Additionally computes max eigenvalue and smoothing factor.'''
  print('CFLstar %1.1f, dlim %1.2f, eps %1.2f' % tuple(x0))
  print('Aspect ratio = %d' % Ly)
  # name for saving plots
  namestring = scheme+'_'+precond+'_'+equation+'_AR'+str(Ly)+'.pdf'
  
  # wavenumber vectors
  Nx = 8
  Ny = 8*Ly
  kx = np.zeros(Nx)
  ky = np.zeros(Ny)
  for i in range(Nx):
    kx[i] = -Nx/2+1+i
  for i in range(Ny):
    ky[i] = -Ny/2+1+i
  # phase angle
  thetax = 2*np.pi*kx/Nx
  thetay = 2*np.pi*ky/Ny
  
  g = np.zeros((Nx,Ny))
  g2 = np.zeros((Nx,Ny,4),dtype=complex)
  maxg = 0
  maxsmooth = 0
  for i in range(Nx):
    for j in range(Ny):
      G = amp_factor_numeric(Nx,Ny,kx[i],ky[j],scheme,precond,equation,x0)
      # max eigenvalue of G
      eg,vg = np.linalg.eig(G)
      g2[i,j,:] = eg
      g[i,j] = np.max(np.absolute(eg))
      maxg=max(maxg,g[i,j])
      #Smoothing factor: frequencies -pi until -pi/2 and pi/2 until pi
      if(i<Nx/4 or i>Nx*3/4 or j<Nx/4 or j>Nx*3/4):
        maxsmooth=max(maxsmooth,g[i,j])

  params = {'legend.fontsize': 'x-large',
            'axes.labelsize': 'x-large',
            'axes.titlesize':'x-large',
            'xtick.labelsize':'x-large',
            'ytick.labelsize':'x-large'}
  pylab.rcParams.update(params)

  plt.figure()
  fig = plt.contour(thetax,thetay,g.transpose())
  plt.clabel(fig, inline=1, fontsize=12)
  plt.title('Amplification factor')
  plt.savefig('ampfactor_'+namestring)

  # eigenvalues of g matrix
  g2 = g2.reshape(Nx*Ny*4)
  plt.figure()
  fig = plt.scatter(np.real(g2),np.imag(g2),marker = '.')
  plt.xlim(-1,1)
  plt.ylim(-1,1)
  plt.title('Eigenvalues of G')
  plt.savefig('eigs_'+namestring)
  #plt.show()
  #plt.close()

  return maxg, maxsmooth

####################################################################

def main():
  '''Discrete Fourier Analysis script.
  Required input arguments:
  RK scheme (0:5), preconditioner (0:4), unsteady (0 or 1).
  See help(get_scheme_params) for choices.'''
  
  # Get RK scheme, preconditioner and equation choices from args
  scheme, precond, equation = get_scheme_params(sys.argv)
  print('RK scheme: %s, preconditioner: %s, equation: %s' % (scheme,precond,equation))
  # aspect ratios
  Ly = np.array([1,100,10000])
  
  # CFLstar,dlim,eps
  cflstarval = 3
  dlimval = 0.25
  epsval = 0.5
  cfl = [cflstarval,cflstarval,cflstarval]
  dlim = [dlimval,dlimval,dlimval]
  eps = [epsval,epsval,epsval]
  for i in range(len(Ly)):
    maxg, maxsmooth = discrete_fourier_analysis([cfl[i],dlim[i],eps[i]],Ly[i],scheme,precond,equation)
    print('Amplification factor %1.4f' % maxg)
    print('Smoothing factor %1.4f\n' % maxsmooth)

if __name__ == "__main__":
  sys.exit(main())
